;;; -*- Lisp -*-


(defun ppolar-B1 (R freq) (- (* 2.0 R (cos (in-Hz freq)))))
(defun ppolar-B2 (R) (* R R))

(defInstrument addflts (start duration numfilts driver freqfun freq0 freq1 ampfun amp cosnum 
			inloc degree distance reverb-amount &rest filts)

  ;; numfilts: number of second order pole sections used.
  ;; driver: 1=white noise, 2=sum-of-cosines, 3=inloc (a sound file name)
  ;; filts: list of lists of filter data: '(freq-fun R-fun frq0 frq1 R0 R1 scaler)

  (let* ((beg (floor (* start sampling-rate)))
	 (end (+ beg (floor (* duration sampling-rate))))
	 (rfs (make-array numFilts :element-type 'envelope))
	 (ffs (make-array numFilts :element-type 'envelope))
	 (pps (make-array numFilts :element-type 'smpflt))
	 (scs (make-array numFilts :element-type 'short-float))
	 (loc (make-locsig :degree degree :distance distance :revscale reverb-amount))
	 (rn (if (= driver 1)
		 (make-randh :frequency freq0)))
	 (cn (if (= driver 2)
		 (make-sum-of-cosines :frequency freq0 :cosines cosnum)))
	 (fn (if (= driver 3) (make-readin :file inloc :start-time 0.0)))
	 (insig 0.0)
	 (outsig 0.0)
	 (ampf (make-func :envelope ampfun :scaler amp :start-time start :duration duration))
	 (frqf (make-func :envelope freqfun :scaler (in-Hz (- freq1 freq0)) :start-time start :duration duration)))
    (loop for i in filts and k from 0 below numfilts do
      (let* ((frqfun (nth 0 i))
	     (R0 (nth 4 i))
	     (R1 (nth 5 i))
	     (G (nth 6 i))
	     (F0 (nth 2 i))
	     (F1 (nth 3 i))
	     (B1-bot (ppolar-B1 R0 F0))
	     (B1-top (ppolar-B1 R1 F1))
	     (B2-bot (ppolar-B2 R0))
	     (B2-top (ppolar-B2 R1))
	     (Rfun (nth 1 i)))
	(setf (aref pps k) (make-ppolar R0 F0))
	(setf (aref ffs k) (make-func :envelope frqfun :offset B1-bot :scaler (- B1-top B1-bot)
				      :start-time start :duration duration))
	(setf (aref rfs k) (make-func :envelope rfun :offset B2-bot :scaler (- B2-top B2-bot)
				      :start-time start :duration duration))
	(setf (aref scs k) G)))
    (Run
     (loop for i from beg to end do
       (setf outsig 0.0)
       (setf insig (* (env ampf) (if (= driver 1) 
				     (randh rn (env frqf))
				   (if (= driver 2)
				       (sum-of-cosines cn (env frqf))
				     (readin fn)))))
       (dotimes (k numFilts)
	 (incf outsig (ppolar (aref pps k) (* (aref scs k) insig)))
	 (setf (smpflt-b1 (aref pps k)) (env (aref ffs k)))
	 (setf (smpflt-b2 (aref pps k)) (env (aref rfs k))))
       (locsig loc i outsig)))))
#|
  (with-sound ()
     (addflts 0 1.0 2 2 '(0 0 100 1) 100 120 '(0 0 50 1 100 0) .25 30 0 0 1.0 0.01
	      '((0 0 100 1) (0 0 100 1) 1200 2400 .9 .995 .005)
	      '((0 0 100 1) (0 0 100 1) 600 1200 .95 .995 .01)))



Date:     Wed, 11 Sep 91 15:40:00 MET DST
From: Rick Taube <taube@ira.uka.de>
To: bil@ccrma.Stanford.EDU
Cc: taube@ira.uka.de
Subject:  addflts question (from Mesias Maiguashca)

; The file addflt.lisp was compiled to the file addflt.fasl.
; The way I understand the instrument is that R0 and R1 give somehow the
; bandwith of the filter. When R0 and R1 have the same value then 
; everything is OK. However when the data shown below is given,
; the bandwith seems to become smaller during 1/2 of the duration,
; (which is OK) and then a gliss. downwards is heard, which I cannot explain.

  
(with-sound ()
     (addflts    
      0.0                  ; start
      5.0                  ; duration
      1                    ; numfilters
      1                    ; driver 1= white noise
     '(0 0 100 0 )         ; freqfun        no function
      10000                ; freq0
      10000                ; freq1
      '(0 0  1 1  90 1 100 0)    ; ampfun
      .1                   ; amp
      30                   ; cosnum, number of cosines
      0                    ; inloc
      0                    ; degree
      1.0                  ; distance
      0.01                 ; reverb amount
      ;  freqfun         R-fun            Frq0   Frq1    R0    R1   AmpScaler
      '((0 0 100 1)    (0 0 50 1 100 1)  440.0   440.0  .5   .999   .1)))
 
; ============================================================
; ZKM, Karlsruhe (mesias maiguashca)

Date: Wed, 11 Sep 91 07:36:17 GMT-0700
From: bil (To: Rick Taube <taube@ira.uka.de>)
To: Rick Taube <taube@ira.uka.de>
Subject: Re: addflts question (from Mesias Maiguashca)

I believe this addflts oddity is a sort of bug in addflts.  I'm
no filter expert (addflts was a translation of an instrument
written by Xavier Serra), but I think the problem is that the
"freqfun" and "Rfun" don't take each other into account.  So
even if the Frq0 and Frq1 are the same, so that freqfun should
be a no-op, if R0 and R1 are different, that gets factored into
the freqfun scaler (it is trying to get the result right, but
ignores the fact that R0 and R1 may not be what it thinks they
are because R-fun is changing them in unnoticed ways).  The
glissando is then an artifact of the very narrow bandwidth of
the filter (R=.999) and the confusion between R-fun and freqfun
(the latter is trying to end up at 440Hz, which is approximately
the end of the glissando, but calculates it's shape without 
noticing that R0 and R1 are getting to their endpoint faster
than expected).

I hesitate to change addflts, because it has been used in this
form for many years.


|#
